from __future__ import annotations

import dataclasses
import errno
import pathlib
import typing

import pysen
from pysen.component import LintComponentBase
from pysen.lint_command import LintCommandBase
from pysen.path import change_dir, resolve_path
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.setting import SettingBase, SettingFile, to_dash_case
from pysen.source import PythonFileFilter

from .preset import UNFIXABLE, preset_dict

LINT_CHECK_COMMAND = ["ruff", "check", "--output-format=full"]
LINT_FIX_COMMAND = ["ruff", "check", "--fix-only", "--show-fixes", "--exit-zero"]
FORMAT_CHECK_COMMAND = ["ruff", "format", "--check", "--diff"]
FORMAT_FIX_COMMAND = ["ruff", "format"]


class RuffCommand(LintCommandBase):
    def __init__(
        self: RuffCommand,
        paths: PathContext,
        source: pysen.Source,
        *,
        inplace_edit: bool,
    ) -> None:
        super().__init__(paths.base_dir, source)
        self._inplace_edit = inplace_edit
        self._setting_path = resolve_path(paths.settings_dir, "pyproject.toml")

    @property
    def name(self: RuffCommand) -> str:
        return "ruff"

    @property
    def has_side_effects(self: RuffCommand) -> bool:
        return self._inplace_edit

    def __call__(self: RuffCommand, reporter: Reporter) -> int:
        sources = self._get_sources(reporter, PythonFileFilter)
        msg = f"Checking {len(sources)} files"
        reporter.logger.info(msg)
        files = [str(s) for s in sources]
        format_command = (
            FORMAT_FIX_COMMAND if self._inplace_edit else FORMAT_CHECK_COMMAND
        )
        lint_command = LINT_FIX_COMMAND if self._inplace_edit else LINT_CHECK_COMMAND
        commands = [lint_command, format_command]

        code = 0
        with change_dir(self.base_dir):
            for command in commands:
                command_result = self._execute_command(command, files, reporter)
                if command_result != 0:
                    code = command_result

        return code

    def _execute_command(
        self: RuffCommand,
        command: typing.List[str],
        files: typing.List[str],
        reporter: Reporter,
    ) -> int:
        chunks = [files]
        code = 0
        while len(chunks) > 0:
            chunk = chunks.pop()
            if len(chunk) == 0:
                continue
            try:
                returncode, _, _ = pysen.process_utils.run(
                    pysen.process_utils.add_python_executable(*command, *chunk),
                    reporter,
                )
                if returncode != 0:
                    code = returncode
            except OSError as e:
                if len(chunk) >= 2 and e.errno == errno.E2BIG:
                    c = len(chunk) // 2
                    chunks.append(chunk[:c])
                    chunks.append(chunk[c:])
                else:
                    raise

        return code


class RuffComponent(LintComponentBase):
    def __init__(
        self: RuffComponent,
        setting: RuffSetting,
        source: pysen.Source,
    ) -> None:
        super().__init__("ruff", source)
        self._setting = setting

    def export_settings(
        self: RuffComponent,
        paths: PathContext,
        files: typing.DefaultDict[str, SettingFile],
    ) -> None:
        files["pyproject.toml"].set_section(*self._setting.export())

    @property
    def targets(self: RuffComponent) -> typing.List[str]:
        return ["lint", "format"]

    def create_command(
        self: RuffComponent,
        target: str,
        paths: PathContext,
        options: pysen.RunOptions,
    ) -> pysen.CommandBase:
        return RuffCommand(paths, self.source, inplace_edit=(target == "format"))


class RuffPlugin(pysen.PluginBase):
    def load(
        self: RuffPlugin,
        file_path: pathlib.Path,
        config_data: pysen.PluginConfig,
        root: pysen.Config,
    ) -> typing.List[RuffComponent]:
        source = pysen.Source(includes=["."])
        preset = "basic"
        target_version = None
        line_length = None
        isort_setting: typing.Union[RuffLintIsortSetting | typing.Dict] = {}

        if config_data.config:
            preset = config_data.config.get("preset", preset)
        if root.lint:
            if root.lint.py_version:
                target_version = root.lint.py_version.short_representation
            if root.lint.line_length:
                line_length = root.lint.line_length
            if root.lint.isort_known_first_party:
                known_first_party = root.lint.isort_known_first_party
                isort_setting = RuffLintIsortSetting(
                    known_first_party=known_first_party
                )
            if root.lint.source:
                source = root.lint.source

        select, ignore = preset_dict[preset]
        setting = RuffSetting(
            target_version=target_version,
            line_length=line_length,
            lint=RuffLintSetting(
                select=select,
                ignore=ignore,
                unfixable=UNFIXABLE,
                isort=isort_setting,
            ),
        )
        return [RuffComponent(setting, source=source)]


@dataclasses.dataclass
class RuffLintSetting:
    select: typing.List[str]
    ignore: typing.List[str]
    unfixable: typing.List[str]
    isort: typing.Union[RuffLintIsortSetting, typing.Dict]


@dataclasses.dataclass
class RuffLintIsortSetting:
    known_first_party: typing.List[str]


@dataclasses.dataclass
class RuffSetting(SettingBase):
    lint: RuffLintSetting
    target_version: typing.Optional[str] = None
    line_length: typing.Optional[int] = 88
    isort: typing.Optional[typing.Dict[str, typing.List[str]]] = None

    def export(
        self: RuffSetting,
    ) -> typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]:
        return ["tool", "ruff"], self.asdict(omit_none=True, naming_rule=to_dash_case)



import pathlib
from typing import List, Optional, Sequence

from pysen.command import CommandBase
from pysen.component import ComponentBase
from pysen.ext.mypy_wrapper import MypyTarget
from pysen.mypy import MypyCommand
from pysen.runner_options import PathContext, RunOptions


class Mypy(ComponentBase):
    def __init__(
        self,
        name: str = "mypy",
        mypy_targets: Optional[Sequence[MypyTarget]] = None,
        settings_dir: Optional[pathlib.Path] = None,
    ) -> None:
        self._name = name
        self._mypy_targets = list(mypy_targets or [])
        self._settings_dir = settings_dir or pathlib.Path(".")

    @property
    def name(self) -> str:
        return self._name

    @property
    def mypy_targets(self) -> List[MypyTarget]:
        return self._mypy_targets

    @property
    def targets(self) -> Sequence[str]:
        return ["lint"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        paths = PathContext(base_dir=paths.base_dir, settings_dir=self._settings_dir)
        if target == "lint":
            return MypyCommand(
                self.name, paths, self.mypy_targets, options.require_diagnostics
            )

        raise AssertionError(f"unknown {target}")



import dataclasses
import errno
import pathlib
from typing import Sequence

import dacite

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    Source,
    git_utils,
    process_utils,
)
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.types import TargetName


def prettier() -> PluginBase:
    return PrettierPlugin()


@dataclasses.dataclass
class PrettierConfig:
    extensions: Sequence[str]
    strict_extentions: bool = False


class PrettierPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        prettier_config = dacite.from_dict(
            PrettierConfig,
            config.config or {},
            config=dacite.Config(strict=True),
        )
        assert isinstance(prettier_config, PrettierConfig)

        return [PrettierComponent(prettier_config, source)]


class PrettierComponent(ComponentBase):
    def __init__(self, config: PrettierConfig, source: Source) -> None:
        self._config = config
        self._source = source

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return PrettierCommand(
            self._config, paths.base_dir, self._source, target == "format"
        )


class PrettierCommand(CommandBase):
    def __init__(
        self,
        config: PrettierConfig,
        base_dir: pathlib.Path,
        source: Source,
        inplace_edit: bool,
    ) -> None:
        self._config = config
        self._base_dir = base_dir
        self._source = source
        self._inplace_edit = inplace_edit

    @property
    def name(self) -> str:
        return "prettier"

    @property
    def has_side_effects(self) -> bool:
        return self._inplace_edit

    def __call__(self, reporter: Reporter) -> int:
        def extention_check(file_path: pathlib.Path) -> bool:
            return file_path.suffix in self._config.extensions

        sources = [
            file_path
            for file_path in self._source.resolve_files(
                self._base_dir,
                extention_check,
                git_utils.check_git_available(self._base_dir),
                reporter,
            )
        ]

        if self._config.strict_extentions:
            sources = list(filter(extention_check, sources))
        str_sources = [str(s) for s in sources]

        command = ["prettier"]
        if self._inplace_edit:
            command.append("--write")
        else:
            command.append("--check")

        code = 0
        chunks = [str_sources]
        while len(chunks) > 0:
            chunk = chunks.pop()
            if len(chunk) == 0:
                continue
            try:
                returncode, _, _ = process_utils.run(command + chunk, reporter)
                if returncode != 0:
                    code = returncode
            except OSError as e:
                if len(chunk) >= 2 and e.errno == errno.E2BIG:
                    c = len(chunk) // 2
                    chunks.append(chunk[:c])
                    chunks.append(chunk[c:])
                else:
                    raise

        return code



from typing import Sequence

from pysen import CommandBase, ComponentBase, RunOptions, Source
from pysen.runner_options import PathContext
from pysen.types import TargetName

from .command import PylintCommand


class PylintComponent(ComponentBase):
    def __init__(self, source: Source) -> None:
        self._source = source

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return PylintCommand(paths.base_dir, self._source, paths.settings_dir)



import logging
import pathlib
from typing import Optional, Sequence

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    SingleFileFormatCommandBase,
    Source,
    process_utils,
)
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.types import TargetName


def shfmt() -> PluginBase:
    return ShfmtPlugin()


class ShfmtPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])
        return [ShfmtComponent(source)]


class ShfmtComponent(ComponentBase):
    def __init__(self, source: Source) -> None:
        self._source = source

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return ShfmtCommand(paths.base_dir, self._source, target == "format")


class ShfmtCommand(SingleFileFormatCommandBase):
    @property
    def name(self) -> str:
        return "shfmt"

    @property
    def has_side_effects(self) -> bool:
        return self.inplace_edit

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.suffix in {".bash", ".sh"}

    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        returncode, stdout, _ = process_utils.run(
            ("shfmt", str(file_path)),
            reporter,
            stdout_loglevel=logging.NOTSET,
        )
        if returncode == 0:
            return stdout
        else:
            return None



import pathlib

from setuptools import find_packages, setup

BASE_DIR = pathlib.Path(__file__).resolve().parent
exec((BASE_DIR / "pysen_plugins/_version.py").read_text())


setup(
    name="pysen-plugins",
    version=__version__,  # type: ignore[name-defined]  # NOQA: F821
    packages=find_packages(),
    description="Collection of pysen plugins",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Toru Ogawa, Ryo Miyajima, Linsho Kaku",
    author_email="ogawa@preferred.jp, ryo@preferred.jp, linsho@preferred.jp",
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Unix",
    ],
    install_requires=["pysen>=0.11.0,<0.12.0"],
    package_data={"pysen_plugins": ["py.typed"]},
)
